"""genshinhelper.
Automatically get Genshin Impact daily check-in rewards from miHoYo bbs and HoYoLAB Community.

GitHub:
    https://github.com/y1ndan/genshinhelper
"""
import json
import time
import uuid
from functools import wraps

from . import journal
from .config import config
from .exceptions import CookiesExpired
from .utils import MESSAGE_TEMPLATE, log, request, get_ds


def _data_handler(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if 'Yuanshen' in self.__class__.__name__:
            self.get_roles()
            roles = self.ROLES_INFO
            if roles and (roles.get('retcode', 1) != 0 or not roles.get('data')):
                raise CookiesExpired(roles.get('message', 'Empty roles data'))
            role_list = roles.get('data', {}).get('list', [])
            if roles and not role_list:
                raise Exception(roles.get('message', 'Empty roles list'))
            if role_list:
                result_list = []
                log.info(f'当前账号绑定了 {len(role_list)} 个角色')
                for role in role_list:
                    # cn_gf01:  天空岛
                    # cn_qd01:  世界树
                    region = role.get('region', 'cn')
                    region_name = role.get('region_name', 'CN')
                    uid = role.get('game_uid', 123456789)
                    self.INFO_URL = self._INFO_URL.format(region, self.ACT_ID, uid)
                    post_data = {
                        'act_id': self.ACT_ID,
                        'region': region,
                        'uid': uid
                    }
                    self._sign_data.update({
                        'region': region,
                        'region_name': region_name,
                        'uid': uid,
                        'post_data': post_data
                    })
                    result_list.append(func(self, *args, **kwargs))
                return ''.join(result_list)
        elif 'Genshin' in self.__class__.__name__:
            try:
                self._sign_data['uid'] = self._cookie.split('account_id=')[
                    1].split(';')[0]
            except Exception as e:
                raise Exception(e)
            self._sign_data['post_data'] = {'act_id': self.ACT_ID}
            return func(self, *args, **kwargs)

    return wrapper


class __BaseCheckin(object):
    USER_AGENT = ''
    REFERER_URL = ''
    ROLES_INFO = ''
    INFO_URL = ''
    REWARD_URL = ''
    SIGN_URL = ''

    def __init__(self, cookie: str = None):
        self._cookie = cookie
        self._sign_data = {}

    def get_header(self):
        header = {
            'Cookie': self._cookie,
            'User-Agent': self.USER_AGENT,
            'Referer': self.REFERER_URL,
            'Accept-Encoding': 'gzip, deflate, br'
        }
        return header

    @property
    def _sign_info(self):
        try:
            response = request(
                'get', self.INFO_URL, headers=self.get_header()).json()
        except Exception as e:
            raise Exception(e)
        return response

    @property
    def _rewards_info(self):
        try:
            response = request(
                'get', self.REWARD_URL, headers=self.get_header()).json()
        except Exception as e:
            raise Exception(e)
        return response

    @_data_handler
    def run(self):
        sign_info = self._sign_info
        rewards_info = self._rewards_info

        today = sign_info.get('data', {}).get('today', '1970-01-01')
        total_sign_day = sign_info.get('data', {}).get(
            'total_sign_day', 0)
        is_sign = sign_info.get('data', {}).get('is_sign')
        first_bind = sign_info.get('data', {}).get('first_bind')
        awards = rewards_info.get('data', {}).get('awards', [])
        region_name = self._sign_data.get('region_name', 'Global')
        uid = self._sign_data.get('uid', 123456789)
        hidden_uid = str(uid).replace(str(uid)[3:-3], '***', 1)
        data = self._sign_data.get('post_data', {})
        region = self._sign_data.get('region')
        travel_notes = 'Olah! Odomu'
        if region in ['cn_gf01', 'cn_qd01']:
            ledger = journal.get_ledger(self._cookie, uid, region)
            travel_notes = '''旅行者 {month} 月札记
    💠原石: {month_primogems}
    🌕摩拉: {month_mora}'''.format(**ledger)

        message = {
            'today': today,
            'region_name': region_name,
            'uid': hidden_uid,
            'total_sign_day': total_sign_day,
            'travel_notes': travel_notes,
            'end': ''
        }

        log.info(f'准备为旅行者 {hidden_uid} 签到...')
        time.sleep(5)

        if is_sign:
            message['award_name'] = awards[total_sign_day - 1].get('name')
            message['award_cnt'] = awards[total_sign_day - 1].get('cnt')
            message['status'] = '👀 旅行者, 你已经签到过了哦'

            # return ''.join(self.message.format(**message))
            return self.message.format(**message)
        else:
            message['award_name'] = awards[total_sign_day].get('name')
            message['award_cnt'] = awards[total_sign_day].get('cnt')
        if first_bind:
            message['status'] = '💪 旅行者, 请先手动签到一次'

            return self.message.format(**message)

        try:
            response = request('post', self.SIGN_URL, headers=self.get_header(
            ), data=json.dumps(data, ensure_ascii=False)).json()
        except Exception as e:
            raise Exception(e)
        # 0:      success
        # -5003:  already checked in
        code = response.get('retcode', 99999)
        if code != 0:
            return response
        message['total_sign_day'] = total_sign_day + 1
        message['status'] = response.get('message')

        log.info('签到完毕')
        return self.message.format(**message)

    @property
    def message(self):
        return MESSAGE_TEMPLATE


class YuanshenCheckin(__BaseCheckin):
    ACT_ID = 'e202009291139501'
    USER_AGENT = 'Mozilla/5.0 (iPhone; CPU iPhone OS 14_0_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) miHoYoBBS/{}'
    REFERER_URL = 'https://webstatic.mihoyo.com/bbs/event/signin-ys/index.html?bbs_auth_required={}&act_id={}&utm_source={}&utm_medium={}&utm_campaign={}'.format('true', ACT_ID, 'bbs', 'mys', 'icon')
    ROLES_INFO = {}
    INFO_URL = ''
    ROLE_URL = 'https://api-takumi.mihoyo.com/binding/api/getUserGameRolesByCookie?game_biz={}'.format('hk4e_cn')
    _INFO_URL = 'https://api-takumi.mihoyo.com/event/bbs_sign_reward/info?region={}&act_id={}&uid={}'
    REWARD_URL = 'https://api-takumi.mihoyo.com/event/bbs_sign_reward/home?act_id={}'.format(ACT_ID)
    SIGN_URL = 'https://api-takumi.mihoyo.com/event/bbs_sign_reward/sign'

    def get_header(self):
        client_type, app_version, ds = get_ds('5')
        header = super().get_header()
        header.update({
            'User-Agent': self.USER_AGENT.format(app_version),
            'x-rpc-device_id': str(uuid.uuid3(uuid.NAMESPACE_URL, self._cookie)).replace(
                '-', '').upper(),
            'x-rpc-client_type': client_type,
            'x-rpc-app_version': app_version,
            'DS': ds
        })
        return header

    def get_roles(self):
        try:
            self.ROLES_INFO = request(
                'get', self.ROLE_URL, headers=super().get_header()).json()
        except Exception as e:
            raise Exception(e)
        return self.ROLES_INFO


class GenshinCheckin(__BaseCheckin):
    ACT_ID = 'e202102251931481'
    USER_AGENT = 'Mozilla/5.0 (iPhone; CPU iPhone OS 14_0_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148'
    REFERER_URL = 'https://webstatic-sea.mihoyo.com/ys/event/signin-sea/index.html?act_id={}'.format(ACT_ID)
    INFO_URL = 'https://hk4e-api-os.mihoyo.com/event/sol/info?lang={}&act_id={}'.format(config.LANGUAGE, ACT_ID)
    REWARD_URL = 'https://hk4e-api-os.mihoyo.com/event/sol/home?lang={}&act_id={}'.format(config.LANGUAGE, ACT_ID)
    SIGN_URL = 'https://hk4e-api-os.mihoyo.com/event/sol/sign?lang={}'.format(config.LANGUAGE)



"""Configuration."""

import json
import os


class Config(object):
    """
    Get all configuration from the config.json file.
    If config.json dose not exist, use config.example.json file.

        Note:   Environment variables have a higher priority,
                if you set a environment variable in your system,
                that variable in the config.json file will be invalid.
    """

    def __init__(self):
        # Open and read the config file
        # project_path = os.path.dirname(os.path.dirname(__file__))
        project_path = os.path.dirname(__file__)
        config_file = os.path.join(project_path, 'config', 'config.json')
        if not os.path.exists(config_file):
            config_file = os.path.join(project_path, 'config', 'config.example.json')

        with open(config_file, 'r', encoding='utf-8') as f:
            self.config_json = json.load(f)

        # Language
        self.LANGUAGE = self.get_config('LANGUAGE')

        # Cookie configs
        # Cookie from https://bbs.mihoyo.com/ys/
        self.COOKIE_MIHOYOBBS = self.get_config('COOKIE_MIHOYOBBS')
        self.COOKIE_MIYOUBI = self.get_config('COOKIE_MIYOUBI')

        # Cookie from https://webstatic-sea.mihoyo.com/ys/event/signin-sea/index.html?act_id=e202102251931481&lang=en-us
        self.COOKIE_HOYOLAB = self.get_config('COOKIE_HOYOLAB')

        # Parameters from Sina Weibo intl. app.   aid=xxx; gsid=xxx; s=xxx
        self.COOKIE_WEIBO = self.get_config('COOKIE_WEIBO')

        # Cookie from https://m.weibo.cn
        self.COOKIE_KA = self.get_config('COOKIE_KA')

        # Notifier configs
        # iOS Bark App
        self.BARK_KEY = self.config_json.get('BARK_KEY')
        if os.environ.get('BARK_KEY'):
            # Customed server
            if os.environ['BARK_KEY'].find(
                    'https') != -1 or os.environ['BARK_KEY'].find('http') != -1:
                self.BARK_KEY = os.environ['BARK_KEY']
            else:
                self.BARK_KEY = f"https://api.day.app/{os.environ['BARK_KEY']}"
        # Official server
        elif self.BARK_KEY and self.BARK_KEY.find('https') == -1 and self.BARK_KEY.find('http') == -1:
            self.BARK_KEY = f'https://api.day.app/{self.BARK_KEY}'

        self.BARK_SOUND = self.get_config('BARK_SOUND')

        # Cool Push
        self.COOL_PUSH_SKEY = self.get_config('COOL_PUSH_SKEY')
        self.COOL_PUSH_MODE = self.get_config('COOL_PUSH_MODE')

        # Custom Notifier Config
        self.CUSTOM_NOTIFIER = self.get_config('CUSTOM_NOTIFIER')

        # DingTalk Bot
        self.DD_BOT_TOKEN = self.get_config('DD_BOT_TOKEN')
        self.DD_BOT_SECRET = self.get_config('DD_BOT_SECRET')

        # Discord webhook
        self.DISCORD_WEBHOOK = self.get_config('DISCORD_WEBHOOK')

        # iGot
        self.IGOT_KEY = self.get_config('IGOT_KEY')

        # pushplus
        self.PUSH_PLUS_TOKEN = self.get_config('PUSH_PLUS_TOKEN')
        self.PUSH_PLUS_USER = self.get_config('PUSH_PLUS_USER')

        # Server Chan
        self.SCKEY = self.get_config('SCKEY')
        self.SCTKEY = self.get_config('SCTKEY')

        # Telegram Bot
        self.TG_BOT_API = self.get_config('TG_BOT_API')
        self.TG_BOT_TOKEN = self.get_config('TG_BOT_TOKEN')
        self.TG_USER_ID = self.get_config('TG_USER_ID')

        # WeChat Work App
        self.WW_ID = self.get_config('WW_ID')
        self.WW_APP_SECRET = self.get_config('WW_APP_SECRET')
        self.WW_APP_USERID = self.get_config('WW_APP_USERID')
        self.WW_APP_AGENTID = self.get_config('WW_APP_AGENTID')

        # WeChat Work Bot
        self.WW_BOT_KEY = self.get_config('WW_BOT_KEY')

    def get_config(self, name: str):
        value = os.environ[name] if os.environ.get(name) else self.config_json.get(name, '')
        return value


config = Config()



from .exceptions import GenshinHelperException
from .utils import request


def get_ledger(cookie, bind_uid, bind_region, month=0):
    url = 'https://hk4e-api.mihoyo.com/event/ys_ledger/monthInfo'
    payload = {
        'month': month,
        'bind_uid': bind_uid,
        'bind_region': bind_region,
        'bbs_presentation_style': 'fullscreen',
        'bbs_auth_required': True,
        'mys_source': 'GameRecord'
    }
    headers = {
        'Referer': 'https://webstatic.mihoyo.com/ys/event/e20200709ysjournal/index.html?bbs_presentation_style=fullscreen&bbs_auth_required=true&mys_source=GameRecord',
        'User-Agent': 'Mozilla/5.0 (iPhone; CPU iPhone OS 14_0_1 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) miHoYoBBS/2.8.0',
        'Accept-Encoding': 'gzip, deflate, br',
        'Cookie': cookie
    }
    response = request('get', url, headers=headers, params=payload).json()
    if response.get('retcode') != 0:
        return {
            'month': response.get('message'),
            'day_primogems': -1,
            'day_mora': -1,
            'month_primogems': -1,
            'month_mora': -1
        }

    day_data = response.get('data', {}).get('day_data', {})
    month_data = response.get('data', {}).get('month_data', {})
    month = response.get('data', {}).get('data_month', -1)
    day_primogems = day_data.get('current_primogems', -1)
    day_mora = day_data.get('current_mora', -1)
    month_primogems = month_data.get('current_primogems', -1)
    month_mora = month_data.get('current_mora', -1)
    ledger = {
        'month': month,
        'day_primogems': day_primogems,
        'day_mora': day_mora,
        'month_primogems': month_primogems,
        'month_mora': month_mora
    }
    return ledger




from .basenotifier import BaseNotifier as Base
from ..config import config


class TelegramBot(Base):
    def __init__(self):
        self.name = 'Telegram Bot'
        self.token = config.TG_BOT_TOKEN if config.TG_BOT_TOKEN and config.TG_USER_ID else ''
        self.retcode_key = 'ok'
        self.retcode_value = 'error_code'

    def send(self, text=Base.app, status=Base.status, desp=Base.desp):
        url = f'https://{config.TG_BOT_API}/bot{config.TG_BOT_TOKEN}/sendMessage'
        data = {
            'chat_id': config.TG_USER_ID,
            'text': f'{text} {status}\n\n{desp}',
            'disable_web_page_preview': True
        }
        return self.push('post', url, data=data)



"""Utilities."""

import hashlib
import logging
import random
import string
import time

import requests

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(levelname)s %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S')

log = logger = logging


def request(method: str,
            url: str,
            max_retries: int = 2,
            params=None,
            data=None,
            json=None,
            headers=None,
            **kwargs):
    # The first HTTP(S) request is not a retry, so need to + 1
    total_requests = max_retries + 1
    for i in range(total_requests):
        try:
            response = requests.Session().request(
                method,
                url,
                params=params,
                data=data,
                json=json,
                headers=headers,
                timeout=21,
                **kwargs)
        except Exception as e:
            log.error(f'Request failed: {url}\n{e}')
            if i == max_retries:
                raise Exception(f'Request failed ({i + 1}/{total_requests}):\n{e}')

            seconds = 5
            log.info(f'Trying to reconnect in {seconds} seconds ({i + 1}/{max_retries})...')
            time.sleep(seconds)
        else:
            return response


def get_cookies(cookies: str = None):
    if '#' in cookies:
        return cookies.split('#')
    else:
        return cookies.splitlines()


def extract_cookie(name: str, cookie: str):
    if name not in cookie:
        raise Exception('Failed to extract cookie: '
            f'The cookie does not contain the `{name}` field.')
    extract_cookie = cookie.split(f'{name}=')[1].split(';')[0]
    return extract_cookie


def cookie_to_dict(cookie):
    cookie_dict = dict([l.split('=', 1) for l in cookie.split('; ')])
    return cookie_dict


def hexdigest(text):
    md5 = hashlib.md5()
    md5.update(text.encode())
    return md5.hexdigest()


def get_ds(type: str = None):
    # 1:  ios
    # 2:  android
    # 4:  pc web
    # 5:  mobile web
    if not type or type == '5' or type == 'mobileweb':
        client_type = '5'
        app_version = '2.3.0'
        salt = 'h8w582wxwgqvahcdkpvdhbh2w9casgfl'
    elif type == '2' or type == 'android':
        client_type = '2'
        app_version = '2.8.0'
        salt = 'dmq2p7ka6nsu0d3ev6nex4k1ndzrnfiy'
    i = str(int(time.time()))
    r = ''.join(random.sample(string.ascii_lowercase + string.digits, 6))
    c = hexdigest(f'salt={salt}&t={i}&r={r}')
    ds = f'{i},{r},{c}'
    return (client_type, app_version, ds)


MESSAGE_TEMPLATE = '''
    {today:#^18}
    🔅[{region_name}]{uid}
    今日奖励: {award_name} × {award_cnt}
    本月累签: {total_sign_day} 天
    签到结果: {status}
    {travel_notes}
    {end:#^18}'''



from .basenotifier import BaseNotifier as Base
from ..config import config 


class CustomNotifier(Base):
    def __init__(self):
        self.name = 'Custom Notifier'
        self.conf = config.CUSTOM_NOTIFIER if config.CUSTOM_NOTIFIER else {}
        self.token = self.url = self.conf.get('url')
        self.data = self.conf.get('data')
        self.retcode_key = self.conf.get('retcode_key')
        self.retcode_value = self.conf.get('retcode_value')

    def send(self, text=Base.app, status=Base.status, desp=Base.desp):
        if not self.token:
            return self.push('post', '')

        title = f'{text} {status}'
        if self.conf['merge_title_and_desp']:
            title = f'{text} {status}\n\n{desp}'
        if self.conf['set_data_title'] and self.conf['set_data_sub_title']:
            self.conf['data'][self.conf['set_data_title']] = {self.conf['set_data_sub_title']: title}
        elif self.conf['set_data_title'] and self.conf['set_data_desp']:
            self.conf['data'][self.conf['set_data_title']] = title
            self.conf['data'][self.conf['set_data_desp']] = desp
        elif self.conf['set_data_title']:
            self.conf['data'][self.conf['set_data_title']] = title

        if self.conf['method'].upper() == 'GET':
            return self.push('get', self.url, params=self.data)
        elif self.conf['method'].upper() == 'POST' and self.conf['data_type'].lower(
        ) == 'json':
            return self.push('post', self.url, json=self.data)
        else:
            return self.push('post', self.url, data=self.data)
